import pysam
import numpy as np

class Variant(object):
    """
    Base class for variants. Simply encapsulates pysam.VariantRecord object, 
    but deals with features like VAFs, or ANN fields (if exists) more conveniently.

    Supported annotation type: snpeff
    """
    def __init__(self, e, annotation_type):
        self.alleles = e.alleles
        self.alts = e.alts
        self.chrom = e.chrom
        self.contig = e.contig
        self.filter = e.filter
        self.format = e.format
        self.id = e.id
        self.info = e.info
        self.pos = e.pos
        self.qual = e.qual
        self.ref = e.ref
        self.rid = e.rid
        self.rlen = e.rlen
        self.samples = e.samples
        self.start = e.start
        self.stop = e.stop
        self.annotation_type = annotation_type

        if annotation_type == 'snpeff':
            self.annotations = [SnpEffAnnotation(s) for s in self.info['ANN']]

    def get_refallele(self):
        return self.ref 

    def get_altallele(self):
        """Return alt allele.
        Note that this method returns the leftmost alt alleles of the entry.
        """
        return self.alts[0]

    def get_normal_refcount(self):
        raise NotImplementedError

    def get_normal_altcount(self):
        raise NotImplementedError

    def get_tumor_refcount(self):
        raise NotImplementedError

    def get_tumor_altcount(self):
        raise NotImplementedError

    def get_normal_vaf(self):
        num_normalref, num_normalalt = self.get_normal_refcount(), self.get_normal_altcount()
        num_total = num_normalref + num_normalalt
        if num_total == 0:
            return np.nan
        else:
            return num_normalalt / (num_normalref + num_normalalt)

    def get_tumor_vaf(self):
        num_tumorref, num_tumoralt = self.get_tumor_refcount(), self.get_tumor_altcount()
        num_total = num_tumorref + num_tumoralt
        if num_total == 0:
            return np.nan
        else:
            return num_tumoralt / (num_tumorref + num_tumoralt)

    def filter_annotation(self, func):
        """Only keep annotation that meet the criterion.
        """
        self.annotations = list(filter(func, self.annotations))
        return self.annotations
    
    def __str__(self):
        r, a = self.get_refallele(), self.get_altallele()
        if self.annotation_type is None:
            return f'{self.chrom}:{self.pos} ({r}->{a})'
        elif self.annotation_type == 'snpeff':
            anns = ','.join([ann.gene_symbol + '[' + ann.hgvs_p + ']' for ann in self.annotations])
            return f'{self.chrom}:{self.pos} ({r}->{a}, {anns})'

class StrelkaSomaticVariant(Variant):
    
    def __init__(self, e, annotation_type=None):
        super(StrelkaSomaticVariant, self).__init__(e, annotation_type)

    def get_normal_refcount(self):
        ref = self.get_refallele()
        return self.samples['NORMAL'][f'{ref}U'][0]
    
    def get_normal_altcount(self):
        alt = self.get_altallele()
        return self.samples['NORMAL'][f'{alt}U'][0]
    
    def get_tumor_refcount(self):
        ref = self.get_refallele()
        return self.samples['TUMOR'][f'{ref}U'][0]

    def get_tumor_altcount(self):
        alt = self.get_altallele()
        return self.samples['TUMOR'][f'{alt}U'][0]

class VarscanSomaticVariant(Variant):
    
    def __init__(self, e, annotation_type=None):
        super(VarscanSomaticVariant, self).__init__(e, annotation_type)
        self.somatic_p = e.info['SPV']
        self.germline_p = e.info['GPV']
        self.somatic_status = e.info['SS']

    def get_normal_refcount(self):
        ref = self.get_refallele()
        return self.samples['NORMAL']['RD']
    
    def get_normal_altcount(self):
        alt = self.get_altallele()
        return self.samples['NORMAL']['AD']
    
    def get_tumor_refcount(self):
        ref = self.get_refallele()
        return self.samples['TUMOR']['RD']

    def get_tumor_altcount(self):
        alt = self.get_altallele()
        return self.samples['TUMOR']['AD']

class SnpEffAnnotation(object):
    def __init__(self, annotation_string):
        tokens = annotation_string.split('|')
        
        self.allele = tokens[0]
        self.annotation = tokens[1].split('&')
        self.impact = tokens[2]
        self.gene_symbol = tokens[3]
        self.gene_id = tokens[4]
        self.feature_type = tokens[5]
        self.feature_id = tokens[6]
        self.transcript_biotype = tokens[7]
        self.rank, self.total = map(int, tokens[8].split('/')) if tokens[8] != '' else [0, 0]
        self.hgvs_c = tokens[9]
        self.hgvs_p = tokens[10]
        self.cDNA_pos, self.cDNA_len = map(int, tokens[11].split('/')) if tokens[11] != '' else [0, 0]
        self.CDS_pos, self.CDS_len = map(int, tokens[12].split('/')) if tokens[12] != '' else [0, 0]
        self.protein_pos, self.protein_len = map(int, tokens[13].split('/')) if tokens[13] != '' else [0, 0]
        self.distance_to_feature = int(tokens[14]) if tokens[14] != '' else 0
    
    def __str__(self):
        return 'Variant Annotation\nAllele: %s\nAnnotation: %s\nPutative impact: %s\nGene: %s(%s)\nFeature: %s(%s)\nTranscript type: %s\nDNA-level variant: %s\nProtein-level variant %s\n' \
                % (self.allele, self.annotation, self.impact, self.gene_symbol, self.gene_id, self.feature_type, self.feature_id, self.transcript_biotype, self.hgvs_c, self.hgvs_p)

class VCF(object):
    """
    Base class for Variant Call Format (VCF).
    """
    def __init__(self, fp, annotation_type):
        self.fp = fp
        self.vcf = pysam.VariantFile(fp)
        self.annotation_type = annotation_type
        self.variant = None

    def fetch(self, chrom=None, start=None, end=None, region=None, filter=None):
        """Yields VCF entries lying between given genomic region.
        Without chrom or region, yields all entries.
        """
        for entry in self.vcf.fetch(reference=chrom, start=start, end=end, region=region):
            if filter:
                if entry.filter.keys()[0] == filter:
                    yield self.variant(entry, self.annotation_type)
            else:
                yield self.variant(entry, self.annotation_type)
    
    def fetch_feature(self, feature_id, chrom=None, start=None, end=None, region=None, filter=None):
        """Fetch variants associated with feature_id. Utilizes info.ANN field to fetch specific feature.
        Passing rough genomic region may significantly reduce running time.
        """
        for entry in self.vcf.fetch(reference=chrom, start=start, end=end, region=region):
            if filter and entry.filter.keys()[0] != filter:
                continue
            
            v = self.variant(entry, self.annotation_type)
            annotations = v.annotations 
            if any(ann.feature_id.startswith(feature_id) for ann in annotations):
                v.filter_annotation(func=lambda ann: ann.feature_id.startswith(feature_id))
                yield v

class StrelkaSomaticVCF(VCF):
    """
    Handles somatic VCF file generated by Strelka2.
    """
    def __init__(self, fp, annotation_type):
        super(StrelkaSomaticVCF, self).__init__(fp, annotation_type)
        self.variant = StrelkaSomaticVariant

class VarscanSomaticVCF(VCF):
    """
    Handles somatic VCF file generated by VarScan.
    """
    def __init__(self, fp, annotation_type):
        super(VarscanSomaticVCF, self).__init__(fp, annotation_type)
        self.variant = VarscanSomaticVariant



import pysam

def get_coverages(bam, chrom, start, end, strict=False):
    """Returns numpy array of coverages in given region.
    If strict=True, trim the coverages lying outside the given region.
    Note that the region should be formatted in zero-based coordinate.
    """
    pileup = pysam.AlignmentFile(bam).pileup(chrom, start, end)
    coverage_dict = {col.pos:col.n for col in pileup if start <= col.pos < end}
    
    if strict:
        positions = list(range(start, end))
        return positions, [coverage_dict.get(pos, 0) for pos in positions]
    else:
        positions = list(coverage_dict.keys())
        min_pos = min(min(positions), start)
        max_pos = max(max(positions), end)
        positions = list(range(min_pos, max_pos + 1))
        return positions, [coverage_dict.get(pos, 0) for pos in positions]



import os
import sys
import pybedtools
import warnings

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from collections import defaultdict, Counter
from dohlee import bed as bed_util
from dohlee import util
from dohlee import plot

from dohlee.thread import threaded

def cv(series):
    if series.mean() == 0:
        warnings.warn('Mean of series is 0 when computing coefficient of variance.')

    return series.std() / series.mean()

def _confidence_interval(meth, unmeth):
    """Compute Agresti-Coull confidence interval.
    Reference:
    https://github.com/dpryan79/MethylDackel/blob/084d926c90c5275d5520fa3346feed144d36bf48/svg.c
    """
    n = meth + unmeth
    zz = 10.8275661707 # qnorm(0.9995)^2
    z = 3.2905267315 # qnorm(0.9995)
    n_dot = n + zz
    p_dot = (1.0 / n_dot) * (meth + 0.5 * zz)

    s = z * np.sqrt((p_dot / n_dot) * (1 - p_dot))
    return p_dot - s , p_dot + s

def _get_inclusion_bounds(df):
    num_position = df.Position.max()
    low, high, middle = int(num_position * 0.2), int(num_position * 0.8), num_position // 2

    # Step 1: Get the average methylation of the middle 60% of the array.
    average_methylation = df[(df.Position > low) & (df.Position <= high)]['CpG Methylation (%)'].mean() / 100

    # Step 2: Determine the min/max of the confidence intervals of the middle 60%.
    cis = [_confidence_interval(row.nMethylated, row.nUnmethylated) for i, row in df[(df.Position > low) & (df.Position <= high)].iterrows()]
    min_ci, max_ci = min([l for (l, h) in cis]), max([h for (l, h) in cis])

    # Get lower inclusion bound.
    lower_bound = 0
    for i in range(middle, 0, -1):
        this_df = df[df.Position == i]
        meth, unmeth, value = this_df.nMethylated.values[0], this_df.nUnmethylated.values[0], this_df['CpG Methylation (%)'].values[0] / 100
        ci = _confidence_interval(meth, unmeth)

        if (
            (ci[1] < average_methylation and value < min_ci and abs(value - average_methylation) > 0.05)
            or (ci[0] > average_methylation and value > max_ci and abs(value - average_methylation) > 0.05)
        ):
            lower_bound = i + 1
            break

    # Get upper inclusion bound.
    upper_bound = 0
    for i in range(middle + 1, num_position + 1):
        this_df = df[df.Position == i]
        meth, unmeth, value = this_df.nMethylated.values[0], this_df.nUnmethylated.values[0], this_df['CpG Methylation (%)'].values[0] / 100
        ci = _confidence_interval(meth, unmeth)

        if (
            (ci[1] < average_methylation and value < min_ci and abs(value - average_methylation) > 0.05)
            or (ci[0] > average_methylation and value > max_ci and abs(value - average_methylation) > 0.05)
        ):
            upper_bound = i - 1
            break

    whole_cis = [_confidence_interval(row.nMethylated, row.nUnmethylated) for i, row in df.iterrows()]
    return lower_bound, upper_bound, whole_cis

def methyldackel_extract_bounds():
    """TODO
    """
    mbias_data = pd.read_table(sys.stdin)
    mbias_data['CpG Methylation (%)'] = mbias_data.apply(axis=1, func=lambda row: row.nMethylated / (row.nMethylated + row.nUnmethylated) * 100)

    commands = []
    for ax_ind, strand in enumerate(['OT', 'OB'], 1):
        bounds = []
        tmp = mbias_data[(mbias_data.Strand == strand)]
        for read in tmp.Read.unique():
            this_data = tmp[tmp.Read == read]

            lower_bound, upper_bound, cis = _get_inclusion_bounds(this_data)
            bounds.append(lower_bound)
            bounds.append(upper_bound)

        if len(bounds) == 2:
            for _ in range(2):
                bounds.append(0)
        commands.append('--%s %s' % (strand, ','.join(map(str, bounds))))

    print(' '.join(commands))


def default_get_bin(pos, start, end, bp_extension, outer_bin_size=50, inner_bin_count=100):
    if start <= pos < start + bp_extension[0]:
        x = pos - start
        bin_number = x // outer_bin_size
    elif start + bp_extension[0] <= pos < end - bp_extension[1]:
        binsize = (end - start - sum(bp_extension)) / inner_bin_count
        x = pos - (start + bp_extension[0])
        bin_number = int(x / binsize) + bp_extension[0] // outer_bin_size
    else:
        x = end - pos
        bin_number = x // outer_bin_size + bp_extension[0] // outer_bin_size + inner_bin_count
    return bin_number


def aggregate_methylation_landscapes(
    annotation_bed_fp,
    methylation_bed_fps,
    condition_fp,
    temp_file_dir,
    SAMPLE_NAME_COL,
    CONDITION_COL,
    bp_extension=None,
    POS_COL='Position',
    METH_LEVEL_COL='Methylation level',
    BIN_COL='Bin',
    agg='mean',
    binning_func=None,
    outer_bin_size=50,
    inner_bin_count=100,
    extension_direction='both',
    threads=4,
    threaded_kws={'progress': False},
):
    """Plot averaged methylation values across the regions which are annotated in the bed file.
    
    :param str annotation_bed_fp: Path to a BED file defining the region of interest.
    :param list methylation_bed_fps: A list of paths of BED files defining the meth-level of each CpG position. 
    :param str condition_fp: Path to a two-column file which represents conditions of samples.
    :param str temp_file_dir: Path to a directory to save caches.
    :param str SAMPLE_NAME_COL: Name of the column denoting IDs of samples.
    :param str CONDITION_COL: Name of the column denoting conditions of samples.
    :param int bp_extension: Amount of extension.
    :param str POS_COL: (default='Position') Name of the column denoting CpG positions.
    :param str METH_LEVEL_COL: (default='Methylation level') Name of the column denoting methylation levels.
    :param str BIN_COL: (default='Bin') Name of the column denoting bins.
    :param str agg: (default='mean') Aggregation method. e.g. 'mean', 'var', 'min', 'max', 'cv'
    :param func binning_func: If possible, pass custom function that represents your binning strategy.
    :param int outer_bin_size: Size of each bin in outer extended regions.
    :param int inner_bin_count: Number of bins in inner regions, which are originally defined in the annotation BED file.
    :param str extension_direction: Extend view toward 'both'/'up'/'down' direction.
    :param int threads: Number of threads to use.
    :param dict threaded_kws: Keyword arguments passed to dohlee.threaded function.
    """
    binning_func = default_get_bin  if binning_func is None else binning_func
    agg          = cv               if agg == 'cv'          else agg

    condition = pd.read_table(condition_fp)
    # Sanity check of the condition file.
    assert SAMPLE_NAME_COL in condition.columns.values, \
        'Column representing sample names %s is not in condition file.' % SAMPLE_NAME_COL

    # Sort and extend annotations with specified amount of bp's.
    ann_bed          = pybedtools.BedTool(annotation_bed_fp).sort()
    interval_lengths = [(iv.end - iv.start) for iv in ann_bed]

    # If extension parameter is not given, set it as the most common interval length in BED file.
    most_common_interval_length = util.most_common_value(Counter(interval_lengths))
    if bp_extension is None:
        bp_extension = [most_common_interval_length] * 2
    
    # Extend each of the intervals in given BED file with the amount given.
    extended_ann_bed = bed_util.extend_bed(ann_bed, bp=bp_extension, direction=extension_direction)

    # Prepare parameters for multithreaded calls.
    params = [(bed, extended_ann_bed, temp_file_dir, binning_func, bp_extension, outer_bin_size, inner_bin_count, SAMPLE_NAME_COL, POS_COL, METH_LEVEL_COL, BIN_COL) for bed in methylation_bed_fps]
    # Run multithreaded runs of aggregating (by computing average) methylation levels.
    aggregated_dfs = list(threaded(func=aggregate_intersecting_methylation_levels, params=params, processes=threads, **threaded_kws))
    final_df = pd.concat(aggregated_dfs).merge(condition, on=SAMPLE_NAME_COL, how='inner')

    # Draw plot.
    ax = plot.get_axis(preset='wide')
    plt.style.use('dohlee')

    condition_mean_val_dict = defaultdict(list)
    for (condition, bin_num), data in final_df.groupby([CONDITION_COL, BIN_COL]).agg(agg).iterrows():
        condition_mean_val_dict[condition].append(data[METH_LEVEL_COL])

    # TODO: Allow user to specify order of conditions to be plotted.
    for condition, y in condition_mean_val_dict.items():
        ax.plot(y, lw=0.5, label=condition)

    ax.set_ylabel('Methylation level')
    ax.tick_params(left=True, length=3, width=0.5, axis='y')

    # Thinner spines.
    for spine in ax.spines.values():
        spine.set_linewidth(0.5)

    # Show legends.
    ax.legend(prop={'size': 'large'})
    return ax

def aggregate_intersecting_methylation_levels(
    methylation_bed_fp,
    extended_ann_bed,
    temp_file_dir,
    binning_func,
    bp_extension,
    outer_bin_size,
    inner_bin_count,
    SAMPLE_NAME_COL,
    POS_COL,
    METH_LEVEL_COL,
    BIN_COL,
):
    sample_name = util.strip_ext(methylation_bed_fp)
    temp_file = os.path.join(temp_file_dir, '%s_aggregated.tsv' % sample_name)

    if not os.path.exists(temp_file):
        methylation_bed = pybedtools.BedTool(methylation_bed_fp).sort()
        intersection_bed = methylation_bed.intersect(extended_ann_bed, loj=True) \
                                          .filter(lambda iv: iv.fields[6] != '.') \
                                          .saveas(os.path.join(temp_file_dir, '%s_tmp.bed' % sample_name))

        data = {
            SAMPLE_NAME_COL: [],
            POS_COL: [],
            METH_LEVEL_COL: [],
            BIN_COL: [],
        }

        for interval in intersection_bed:
            # Start and end position of the overlapping annotation interval.
            try:
                start, end = int(interval.fields[7]), int(interval.fields[8])
            except:
                with open(os.path.join(temp_file_dir, 'error.txt'), 'w') as outFile:
                    print(interval, file=outFile)
                raise IndexError 

            data[SAMPLE_NAME_COL].append(sample_name)
            data[POS_COL].append(interval.start)
            data[METH_LEVEL_COL].append(interval.score)
            data[BIN_COL].append(binning_func(interval.start, start, end, bp_extension, outer_bin_size, inner_bin_count))

        data = pd.DataFrame(data)
        data[METH_LEVEL_COL] = data[METH_LEVEL_COL].astype(np.float32)
        data.to_csv(temp_file, sep='\t', index=False, header=True)
    else:
        data = pd.read_table(temp_file)
        assert SAMPLE_NAME_COL in data.columns.values

    aggregated = data.groupby([BIN_COL]).agg({
        SAMPLE_NAME_COL: 'first',
        POS_COL: 'first',
        METH_LEVEL_COL: 'mean',
    })
    aggregated[BIN_COL] = aggregated.index
    return aggregated



import matplotlib
matplotlib.use('Agg')

import sys
import os
import itertools
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import pysam
import numpy as np
import scipy.stats as stats
import seaborn as sns
import dohlee.seq as seq

from adjustText import adjust_text
from collections import Counter, defaultdict
from sklearn.decomposition import PCA
from dohlee import analysis
from matplotlib.lines import Line2D
from collections import Counter
from functools import wraps
from sklearn.manifold import TSNE
from urllib import request
from umap import UMAP
from matplotlib.ticker import ScalarFormatter


def _get_ax_to_draw(ax, figsize=None):
    """If ax is not specified, return an axis to draw a plot.
    Otherwise, return ax.
    """
    return ax or get_axis(figsize=figsize)


def _try_save(file, dpi=300):
    """If file is specified, save the figure to the file with given resolution in dpi.
    Otherwise, show the figure.
    """
    return None if file is None else plt.savefig(file, dpi=dpi)


def save(file, dpi=300, tight_layout=True):
    """Save plot to a file.

    :param str file: Path to the resulting image file.
    :param int dpi: (default=300) Resolution.
    :param bool tight_layout: (default=True) Whether to run plt.tight_layout() before saving the plot.
    """
    if tight_layout:
        plt.tight_layout()
    plt.savefig(file, dpi=dpi)


def clear():
    """Clear the plot.
    """
    plt.clf()
    plt.close()


def _my_plot(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        ax = _get_ax_to_draw(kwargs.get('ax', None), kwargs.get('figsize', None))
        if 'ax' in kwargs:
            del kwargs['ax']
        if 'figsize' in kwargs:
            del kwargs['figsize']

        if 'title' in kwargs:
            ax.set_title(kwargs['title'])
            del kwargs['title']
        if 'xlim' in kwargs:
            ax.set_xlim(kwargs['xlim'])
            del kwargs['xlim']
        if 'ylim' in kwargs:
            ax.set_ylim(kwargs['ylim'])
            del kwargs['ylim']
        if 'xlabel' in kwargs:
            ax.set_xlabel(kwargs['xlabel'])
            del kwargs['xlabel']
        if 'ylabel' in kwargs:
            ax.set_ylabel(kwargs['ylabel'])
            del kwargs['ylabel']

        file_path = kwargs.get('file', None)
        if 'file' in kwargs:
            del kwargs['file']

        rotate_xticklabels = kwargs.get('rotate_xticklabels', 0)
        if 'rotate_xticklabels' in kwargs:
            del kwargs['rotate_xticklabels']

        xticklabels = kwargs.get('xticklabels', True)
        if 'xticklabels' in kwargs:
            del kwargs['xticklabels']

        yticklabels = kwargs.get('yticklabels', True)
        if 'yticklabels' in kwargs:
            del kwargs['yticklabels']

        legend_size = kwargs.get('legend_size', None)
        if 'legend_size' in kwargs:
            del kwargs['legend_size']

        legend_title = kwargs.get('legend_title', None)
        if 'legend_title' in kwargs:
            del kwargs['legend_title']

        ax_result = func(*args, ax=ax, **kwargs)
        if rotate_xticklabels:
            ax_result.set_xticklabels(
                ax_result.get_xticklabels(),
                ha='right',
                rotation=rotate_xticklabels,
            )

        if not xticklabels:
            ax.set_xticks([])
            ax.set_xlabel(ax.get_xlabel(), labelpad=7.0)

        if not yticklabels:
            ax.set_yticks([])
            ax.set_ylabel(ax.get_ylabel(), labelpad=7.0)

        if legend_size is not None or legend_title is not None:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels, prop={'size': legend_size}, title=legend_title)

        _try_save(file_path)
        return ax_result
    return wrapper


def set_suptitle(title):
    """Set suptitle for the plot.
    """
    plt.suptitle(title)


# Set plot preference which looks good to me.
def set_style(style='white', palette='deep', context='talk', font='FreeSans', font_scale=1.00):
    """Set plot preference in a way that looks good to me.
    """
    import matplotlib.font_manager as font_manager
    import cycler

    # Update font list.
    mpl_data_dir = os.path.dirname(mpl.matplotlib_fname())
    font_files = font_manager.findSystemFonts(os.path.join(mpl_data_dir, 'fonts', 'ttf'))
    font_list = font_manager.createFontList(font_files)
    font_manager.fontManager.ttflist.extend(font_list)

    sns.set(
        style=style,
        palette=palette,
        context=context,
        font=font,
        font_scale=font_scale,
    )
    scale = 1.3
    plt.rc('axes', linewidth=1.33, labelsize=14)
    plt.rc('xtick', labelsize=10 * scale)
    plt.rc('ytick', labelsize=10 * scale)

    plt.rc('xtick', bottom=True)
    plt.rc('xtick.major', size=5 * scale, width=1.33)
    plt.rc('xtick.minor', size=5 * scale, width=1.33)

    plt.rc('ytick', left=True)
    plt.rc('ytick.major', size=5 * scale, width=1.33)
    plt.rc('ytick.minor', size=5 * scale, width=1.33)

    plt.rc('legend', fontsize=7 * scale)
    plt.rc('grid', color='grey', linewidth=0.5, alpha=0.33)
    plt.rc('font', family='Helvetica Neue')

    color_palette = [
        '#005AC8',
        '#AA0A3C',
        '#0AB45A',
        '#FA7850',
        '#8214A0',
        '#FA78FA',
        '#A0FA82',
        '#006E82',
        '#00A0FA',
        '#14D2DC',
        '#F0F032',
        '#FAE6BE',
    ]

    mpl.rcParams['axes.prop_cycle'] = cycler.cycler(color=color_palette)


def get_axis(figsize=None, transpose=False, dpi=300):
    """Get plot axis with predefined/user-defined width and height.

    >>> ax = get_axis()
    >>> ax = get_axis(figsize=(7.2, 4.45))

    :param float scale: Figure size scale. Width and height will be scale with this value.
    :param tuple figsize: Use user-defined width and height. If this is given, `scale` parameter will be ignored.
    :param bool transpose: Swap width and height.
    """
    w, h = 7.2, 4.45  # Nature double-column preset inches.
    if figsize is not None:
        w, h = figsize

    if transpose:
        w, h = h, w

    fig = plt.figure(figsize=(w, h))
    ax = fig.add_subplot(111)
    return ax


@_my_plot
def get_grid(shape=(1, 1), ax=None, hspace=None, wspace=None):
    """TODO
    """
    subplotspec = ax.get_subplotspec()
    return subplotspec.subgridspec(*shape, hspace=hspace, wspace=wspace)


def get_axis_from_grid(grid):
    """TODO
    """
    return plt.subplot(grid)


@_my_plot
def frequency(data, order=None, sort_by_values=False, dy=0.01, ax=None, **kwargs):
    """Plot frequency bar chart.

    >>> frequency([1, 2, 2, 3, 3, 3], order=[3, 1, 2], sort_by_values=True)

    :param list data: A list of elements.
    :param list order: A list of elements which represents the order of the elements to be plotted.
    :param bool sort_by_values: If True, the plot will be sorted in decreasing order of frequency values.
    :param float dy: Gap between a bar and its count label.
    :param pyplot-axis ax: Axis to draw the plot.
    """
    counter = Counter(data)
    if order is None:
        if sort_by_values:
            order = sorted(counter, key=counter.get, reverse=True)
        else:
            order = sorted(counter.keys())
    else:
        assert set(order) == set(counter.keys()), 'The order must contain all the elements.'
    counts = [counter[key] for key in order]

    ax = _get_ax_to_draw(ax)

    # Some parameters used for plot configuration.
    height = max(counts) * 1.167
    xticks = list(range(len(order)))

    # Preset plot.
    ax.set_xticks(xticks)
    ax.set_xticklabels(order)
    ax.set_xlim([-0.66, len(counter) - 0.33])
    ax.set_ylim([0, height])

    # Plot bar chart.
    ax.bar(x=xticks,
           height=counts,
           width=0.66,
           **kwargs)

    # Add text indicating frequency for each bar.
    for x, count in zip(xticks, counts):
        ax.text(
            x=x,
            y=count + dy,
            s=str(count),
            va='bottom',
            ha='center',
            fontsize=12,
        )

    return ax


@_my_plot
def histogram(data, ax=None, **kwargs):
    """Draw a histogram.

    >>> histogram(data=data, ax=ax, lw=1.55)

    :param list data: A list containing values. Density of the values will be drawn as a histogram.
    :param axis ax: Matplotlib axis to draw the plot on.
    """
    ax.hist(data, color='black', ec='white', lw=1.33, **kwargs)
    return ax


@_my_plot
def boxplot(data, x, y, hue=None, ax=None, strip=False, **kwargs):
    """Draw a boxplot.

    >>> boxplot(data, x='species', y='sepal_length', strip=True)

    :param dataframe data: Dataframe for boxplot.
    :param str x: Column name representing x variable of the plot.
    :param str y: Column name representing y variable of the plot.
    :param axis ax: (Optional) Matplotlib axis to draw the plot on.
    :param bool strip: (default=False) Draw overlapped stripplot.
    """
    fliersize = 0 if strip else 5
    sns.boxplot(data=data, x=x, y=y, hue=hue, linewidth=1.33, flierprops={'marker': '.'}, fliersize=fliersize, ax=ax, **kwargs)
    if strip:
        sns.stripplot(data=data, x=x, y=y, hue=hue, jitter=.03, color='k', size=5, ax=ax)

    return ax


@_my_plot
def volcano(data, x, y, padj, label, cutoff=0.05, sample1=None, sample2=None, ax=None):
    """Draw a volcano plot.

    >>> volcano(data=data,
                x='log2FoldChange',
                y='pvalue',
                label='Gene_Symbol',
                cutoff=0.05,
                padj='padj',
                figsize=(10.8, 8.4))

    :param dataframe data: A dataframe resulting from DEG-discovery tool.
    :param str x: Column name denoting log2 fold change.
    :param str y: Column name denoting p-value. (Note that p-values will be log10-transformed, so they should not be transformed beforehand.)
    :param str padj: Column name denoting adjusted p-value.
    :param str label: Column name denoting gene identifier.
    :param float cutoff: (Optional) Adjusted p-value cutoff value to report significant DEGs.
    :param str sample1: (Optional) First sample name.
    :param str sample2: (Optional) Second sample name.
    :param axis ax: (Optional) Matplotlib axis to draw the plot on.
    """
    # Set x and y extent.
    x_limit = max(-np.min(data[x].values), np.max(data[x].values))
    x_extent = [-x_limit * 1.22, x_limit * 1.22]
    ax.set_xlim(x_extent)
    ax.set_ylim([0, max(-np.log10(data[y].values)) * 1.1])

    data_not_significant = data[data[padj] >= cutoff]
    ax.scatter(data_not_significant[x].values, -np.log10(data_not_significant[y].values), color='grey', marker='.')

    data_significant = data[data.padj < cutoff]
    ax.scatter(data_significant[x].values, -np.log10(data_significant[y].values), color='red', marker='.')

    texts = [ax.text(row[x], -np.log10(row[y]), row[label], fontsize=12) for _, row in data_significant.iterrows()]
    adjust_text(texts)

    line = Line2D([0], [0], color='red', lw=2.33, label='Adjusted p < %g' % cutoff)
    plt.legend(handles=[line])

    if sample1 is not None and sample2 is not None:
        ax.set_xlabel(r'$log_2$FC ($log_{2}\frac{%s}{%s}$)' % (sample1.replace(' ', '\ '), sample2.replace(' ', '\ ')))
    else:
        ax.set_xlabel(r'$log_2$FC')
    ax.set_ylabel(r'$log_{10}$(p-value)')
    return ax


@_my_plot
def pca(data, labels=None, ax=None, **kwargs):
    '''Draw a simple principle component analysis plot of the data.

    :param matrix data: Input data. Numpy array recommended.
    :param list labels: (Optional) Corresponding labels to each datum. If specified, data points in the plot will be colored according to the label.
    :param axis ax: (Optional) Matplotlib axis to draw the plot on.
    :param kwargs: Any other keyword arguments will be passed onto matplotlib.pyplot.scatter.
    '''
    # Fit PCA and get pc's
    pca = PCA(n_components=2)
    pca.fit(data)
    pc = pca.transform(data)

    if labels is None:
        ax.scatter(x=pc[:, 0], y=pc[:, 1], **kwargs)

    else:
        # If labels are attached, color them in different colors
        labels = np.array(labels)
        for label in set(labels):
            toDraw = (labels == label)  # only draw these points this time

            ax.scatter(x=pc[toDraw, 0], y=pc[toDraw, 1], label=label, **kwargs)
            ax.legend(loc='best')

    # show explained variance ratio in the plot axes
    explainedVarianceRatio = pca.explained_variance_ratio_
    ax.set_xlabel('PC1 ({:.2%})'.format(explainedVarianceRatio[0]))
    ax.set_ylabel('PC2 ({:.2%})'.format(explainedVarianceRatio[1]))
    return ax


@_my_plot
def tsne(data, labels=None, ax=None, **kwargs):
    '''Draw a T-SNE analysis plot of the data.

    :param matrix data: Input data. Numpy array recommended.
    :param list labels: (Optional) Corresponding labels to each datum. If specified, data points in the plot will be colored according to the label.
    :param axis ax: (Optional) Matplotlib axis to draw the plot on.
    :param kwargs: Any other keyword arguments will be passed onto matplotlib.pyplot.scatter.
    '''
    # Fit T-SNE and get embeddings.
    embeddings = TSNE(n_components=2).fit_transform(data)

    if labels is None:
        ax.scatter(x=embeddings[:, 0], y=embeddings[:, 1], **kwargs)

    else:
        # If labels are attached, color them in different colors
        labels = np.array(labels)
        for label in set(labels):
            toDraw = (labels == label)  # only draw these points this time

            ax.scatter(
                x=embeddings[toDraw, 0],
                y=embeddings[toDraw, 1],
                label=label,
                **kwargs
            )
            ax.legend(loc='best')
    return ax

@_my_plot
def coverages(path, chrom, start, end, strict=False, tick_every=1000, ax=None, **kwargs):
    positions, covs = seq.get_coverages(path, chrom, start, end, strict=strict)

    for pos, cov in zip(positions, covs):
        if cov == 0:
            continue
        ax.bar(pos, cov, width=1, color='grey', lw=0)
    
    ticks = [i for i in range(start, end) if i % tick_every == 0]
    ax.set_xticks(ticks)
    ax.ticklabel_format(style='plain', useOffset=False)

    for loc, spine in ax.spines.items():
        if loc == 'left':
            spine.set_position(('outward', 20))
            spine.set_smart_bounds(True)
    
    # Hide spines.
    for d in ['top', 'right']:
        ax.spines[d].set_visible(False)

def has_no_overlap(r1, r2):
    """Returns True if the two reads overlap.
    """
    r1_start, r1_end = r1.reference_start, r1.reference_end
    r2_start, r2_end = r2.reference_start, r2.reference_end

    return (r2_end <= r1_start) or (r1_end <= r2_start)

def plot_bisulfite_read(read, depth_reads_dict, chrom, start, end, strict, ax):
    # Find appropriate depth that this read does not overlap
    # with the other reads.
    s, e = read.reference_start, read.reference_end
    if strict and not (start <= s < end and start <= e < end):
        return depth_reads_dict

    good_depth = 0
    found_good_depth = False
    if len(depth_reads_dict) == 0:
        found_good_depth = True

    for depth in sorted(depth_reads_dict.keys()):
        if all(has_no_overlap(read, r) for r in depth_reads_dict[depth]):
            appropriate_depth = depth
            found_good_depth = True
            break
    
    if not found_good_depth:
        good_depth = max(depth_reads_dict.keys()) + 1
    
    depth_reads_dict[good_depth].append(read)

    # Plot read.
    len_read = read.reference_length
    color = '#33333333'
    rect = patches.Rectangle((s, good_depth), len_read, 1, color=color, lw=0)
    ax.add_patch(rect)
    
    # Plot methylation state.
    # Methylated: red
    # Unmethylated: blue
    for offset, state in enumerate(read.get_tag('XM')):
        if state not in 'zZ':
            continue
        color = ['blue', 'red'][state == 'Z']
        meth_state = patches.Rectangle((s + offset, good_depth), 1, 1, color=color)
        ax.add_patch(meth_state)

    return depth_reads_dict

@_my_plot
def bisulfite(path, chrom, start, end, ax=None,
        tick_every=1000, strict=False, **kwargs):
    depth_reads_dict = defaultdict(list)

    # TODO: Automatically index alignment file, if it is not indexed?
    for read in pysam.AlignmentFile(path, require_index=True).fetch(chrom, start, end):
        depth_reads_dict = plot_bisulfite_read(read, depth_reads_dict, chrom, start, end, strict, ax)

    max_depth = max(col.n for col in pysam.AlignmentFile(path).pileup(chrom, start, end))
    ax.set_xlim(start, end)
    ax.set_ylim(0, max_depth)

    ticks = [i for i in range(start, end) if i % tick_every == 0]
    ax.set_xticks(ticks)
    ax.ticklabel_format(style='plain', useOffset=False)

    # TODO: Modularize.
    # move y axis to left.
    for loc, spine in ax.spines.items():
        if loc == 'left':
            spine.set_position(('outward', 20))
            spine.set_smart_bounds(True)

    # Hide spines.
    for d in ['top', 'right']:
        ax.spines[d].set_visible(False)

@_my_plot
def mutation_signature(data, ax=None, **kwargs):
    def clear_spines(axis):
        # TODO: Modularize.
        # Hide spines.
        directions = ['top', 'right', 'bottom', 'left']
        for d in directions:
            axis.spines[d].set_visible(False)
        # Hide ticklabels.
        axis.set_xticklabels([])
        axis.set_yticklabels([])

    # (context, alt) -> index in 96-dim vector.
    def mut2ind(context, alt):
        ref = context[1]
        assert ref in ['C', 'T'], 'ref not in [C, T], ref: %s' % ref

        if ref == 'C':
            return ('AGT'.find(alt)) * 16 + c_contexts.index(context)
        else:
            return (3 + 'ACG'.find(alt)) * 16 + t_contexts.index(context)

    # Define trinucleotide contexts.
    c_contexts = [p + 'C' + n for (p, n) in itertools.product('ACGT', 'ACGT')]
    t_contexts = [p + 'T' + n for (p, n) in itertools.product('ACGT', 'ACGT')]

    # Convert data(Counter) as a 96-dimension vector.
    vectorized_data = np.zeros(96)
    for (context, alt), count in data.items():
        vectorized_data[mut2ind(context, alt)] = count
    vectorized_data = vectorized_data / vectorized_data.sum()

    #
    # Draw figure.
    #
    subplotspec = ax.get_subplotspec()
    grid = subplotspec.subgridspec(7, 6, hspace=0.03, wspace=0.03)

    # Draw annotation bars above the main plot.
    annotation_axes = [plt.subplot(grid[0, i]) for i in range(6)]
    annotations = ['C>A', 'C>G', 'C>T', 'T>A', 'T>C', 'T>G']
    colors = [
        (30, 190, 240),
        (5, 7, 9),
        (231, 39, 38),
        (202, 202, 202),
        (161, 207, 100),
        (238, 200, 197),
    ]

    for axis, annotation, color in zip(annotation_axes, annotations, colors):
        clear_spines(axis)
        axis.text(x=0.5, y=0.5, s=annotation, ha='center', fontsize='large')

        rect = patches.Rectangle((0, 0), width=1, height=0.33, facecolor=[x/255 for x in color], linewidth=0)
        axis.add_patch(rect)

    # Draw main plot.
    main_axis = plt.subplot(grid[1:, :])

    xticklabel_dx = 0.11
    main_axis.set_xticks(np.arange(0, 96, 1) + xticklabel_dx)
    main_axis.set_xticklabels(c_contexts * 3 + t_contexts * 3, rotation=90, fontsize='large', ha='left', va='top', fontdict={'family': 'Dejavu Sans Mono'})
    main_axis.tick_params(axis='x', pad=3)
    main_axis.set_xlim([0, 96])
    main_axis.spines['top'].set_visible(True)
    main_axis.spines['right'].set_visible(False)
    main_axis.yaxis.set_tick_params(size=5)
    main_axis.set_ylabel('Relative contribution')

    data = np.array([np.random.randint(0, 30) for _ in range(96)])
    data = data / data.sum()

    bars = main_axis.bar(np.arange(0, 96, 1) + 0.5, vectorized_data, width=0.5)
    for i in range(6):
        start, end = 16 * i, 16 * (i+1)
        r, g, b = colors[i]
        for bar in bars[start:end]:
            bar.set_facecolor([r / 255, g / 255, b / 255])


@_my_plot
def linear_regression(x, y, regression=True, ax=None, color='k'):
    """TODO
    """
    # Perform linear regression.
    slope, intercept, r_value, p_value, std_err = stats.linregress(x, y)

    ax.scatter(x, y, s=5, color=color)
    x_extent = np.array(ax.get_xlim())
    ax.plot(x_extent, slope * x_extent + intercept, lw=1, color=color)

    legend = ax.legend(
        labels=['$R^2$ = %.3f, p = %.3g' % (r_value ** 2, p_value)],
        loc='best',
        fontsize='small',
        handlelength=0,
        handletextpad=0
    )
    for item in legend.legendHandles:
        item.set_visible(False)


@_my_plot
def m_bias(mbias_data, ax=None):
    """TODO
    """
    mbias_data['CpG Methylation (%)'] = mbias_data.apply(axis=1, func=lambda row: row.nMethylated / (row.nMethylated + row.nUnmethylated) * 100)

    fig = plt.figure(figsize=(8.4 * 2.2, 8.4), dpi=300)

    commands = []
    for ax_ind, strand in enumerate(['OT', 'OB'], 1):
        bounds = []

        ax = fig.add_subplot(1, 2, ax_ind)
        ax.set_ylim([0, max(70, max(mbias_data['CpG Methylation (%)']))])
        ax.set_title(strand, fontsize='x-large')
        ax.set_xlabel('Position along mapped read')
        ax.set_ylabel('CpG Methylation (%)')

        tmp = mbias_data[(mbias_data.Strand == strand)]
        for read in tmp.Read.unique():
            this_data = tmp[tmp.Read == read]

            x = this_data['Position'].values
            y = this_data['CpG Methylation (%)'].values
            ax.plot(x, y, lw=1, label='Read %d' % read)
            ax.set_xlim([0, max(x)])

            for i in range(0, max(x), 10):
                ax.axvline(x=i, linestyle='--', linewidth=0.66, color='grey', alpha=0.33)

            lower_bound, upper_bound, cis = analysis._get_inclusion_bounds(this_data)
            ax.axvline(x=lower_bound, linestyle='--', linewidth=1, color='red', alpha=0.66)
            ax.axvline(x=upper_bound, linestyle='--', linewidth=1, color='red', alpha=0.66)

            ax.legend(fontsize='large', loc='best')

            bounds.append(lower_bound)
            bounds.append(upper_bound)

        if len(bounds) == 2:
            for _ in range(2):
                bounds.append(0)
        commands.append('--%s %s' % (strand, ','.join(map(str, bounds))))


@_my_plot
def stacked_bar_chart(data, x, y, ax=None, sort=False, reverse=True, sort_by=None, group=None, group_order=None, group_label=True):
    """TODO
    """
    ax = _get_ax_to_draw(ax)
    ax.margins(x=5e-3)
    if not isinstance(y, list):
        y = list(y)

    x_values = data[x].values
    y_values = data[y].values.T
    bottoms = data[y].values.cumsum(axis=1).T
    x_positions = np.arange(len(x_values))

    if group:
        groups = data[group].values
    if sort:
        if sort_by is None:
            # If sort_by parameter is not specified, sort by the sum of the heights of stacked bars.
            order_mask = np.argsort(data[y].values.sum(axis=1))
        else:
            assert sort_by in y, 'Column name in sort_by parameter should appear in y parameter.'
            order_mask = np.argsort(data[sort_by].values)
        if reverse:
            order_mask = order_mask[::-1]
        x_values = x_values[order_mask]
        y_values = y_values[:, order_mask]
        bottoms = bottoms[:, order_mask]
        if group:
            groups = groups[order_mask]

    if group is not None:
        group_mask = []
        group_count = Counter()

        group_order = group_order if group_order is not None else data[group].unique()
        for g in group_order:
            for i, g_tmp in enumerate(groups):
                if g_tmp == g:
                    group_count[g] += 1
                    group_mask.append(i)

        # Sort x_values, y_values and bottoms to reflect group assignments.
        # Convert to list to use `insert` method.
        x_values = list(x_values[group_mask])
        y_values = y_values[:, group_mask]
        bottoms = bottoms[:, group_mask]

        # Insert empty space between each of the two groups.
        insert_index = 0
        group_label_position = []
        for i, g in enumerate(group_order):
            group_label_position.append(insert_index + group_count[g] / 2)
            insert_index += group_count[g]
            x_values.insert(insert_index, '')
            num_rows = y_values.shape[0]
            y_values = np.hstack([y_values[:, :insert_index], np.zeros((num_rows, 1)), y_values[:, insert_index:]])
            bottoms = np.hstack([bottoms[:, :insert_index], np.zeros((num_rows, 1)), bottoms[:, insert_index:]])
            insert_index += 1

        x_positions = np.arange(len(x_values))

    ax.bar(x_positions, y_values[0], label=y[0])
    for i in range(1, len(y_values)):
        ax.bar(x_positions, y_values[i], bottom=bottoms[i-1], label=y[i])

    ax.legend()
    if group is not None and group_label:
        ax.set_xticks(group_label_position)
        ax.set_xticklabels(group_order, ha='center')
    else:
        ax.set_xticks(x_positions)
        ax.set_xticklabels(x_values)

    ax.set_ylim([0, ax.get_ylim()[1]])
    ax.tick_params('y', length=5, width=1, which='major')
    return ax

@_my_plot
def umap(data, labels=None, ax=None, **kwargs):
    '''Draw a UMAP embedding plot of the data.

    :param matrix data: Input data. Numpy array recommended.
    :param list labels: (Optional) Corresponding labels to each datum. If specified, data points in the plot will be colored according to the label.
    :param axis ax: (Optional) Matplotlib axis to draw the plot on.
    :param kwargs: Any other keyword arguments will be passed onto matplotlib.pyplot.scatter.
    '''
    # Apply UMAP and get embeddings.
    reducer = UMAP()
    embeddings = reducer.fit_transform(data)

    if labels is None:
        ax.scatter(x=embeddings[:, 0], y=embeddings[:, 1], **kwargs)

    else:
        # If labels are attached, color them in different colors
        labels = np.array(labels)
        for label in set(labels):
            toDraw = (labels == label)  # only draw these points this time

            ax.scatter(
                x=embeddings[toDraw, 0],
                y=embeddings[toDraw, 1],
                label=label,
                **kwargs
            )
            ax.legend(loc='best')
    return ax


@_my_plot
def dimensionality_reduction(data, labels, ax=None):
    ax = get_axis(figsize=(20, 4))
    grid = get_grid((1, 5), ax=ax)
    axes = [get_axis_from_grid(grid[i]) for i in range(5)]

    common_args = dict(
        data=data,
        labels=labels,
        # FIXME: HARD-CODED legend size.
        legend_size='medium',
        xticklabels=False,
        yticklabels=False,
    )

    pca(ax=axes[0], title='PCA', **common_args)
    tsne(ax=axes[1], title='T-SNE', **common_args)
    umap(ax=axes[2], title='UMAP', **common_args)


@_my_plot
def line(y, data=None, x=None, hue=None, ax=None, agg='mean', **kwargs):
    if data is None:
        if x is None:
            ax.plot(y, **kwargs)
        else:
            ax.plot(x, y, **kwargs)
    else:  # `data` is not None.
        if hue is None:
            if x is None:
                ax.plot(data[y], **kwargs)
            if x is not None:
                ax.plot(data[x], data[y], **kwargs)
        else:
            # `data` is not None, and `hue` is not None.
            # We should aggregate data by `hue` column and plot each line separately.
            if x is None:
                aggregated = data.groupby([hue, x, y]).agg(agg)



import pybedtools
import dohlee.bed as bed_util

from nose.tools import raises

def extend_bed_up_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=10,
        direction='up',
    )

    assert result_bed[0].chrom == 'chr1'
    assert result_bed[0].start == 90
    assert result_bed[0].end == 200


def extend_bed_down_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=10,
        direction='down',
    )

    assert result_bed[0].chrom == 'chr1'
    assert result_bed[0].start == 100
    assert result_bed[0].end == 210


def extend_bed_both_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=10,
        direction='both',
    )

    assert result_bed[0].chrom == 'chr1'
    assert result_bed[0].start == 90
    assert result_bed[0].end == 210


def extend_bed_clipping_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=150,
        direction='both',
    )

    assert result_bed[0].chrom == 'chr1'
    assert result_bed[0].start == 0
    assert result_bed[0].end == 350


@raises(AssertionError)
def extend_bed_invalid_bp_failing_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=[1, 2, 3],
    )


def extend_bed_single_element_list_bp_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=[10],
    )

    assert result_bed[0].chrom == 'chr1'
    assert result_bed[0].start == 90
    assert result_bed[0].end == 210


def extend_bed_double_element_list_bp_test():
    # chr1  100 200
    bed = pybedtools.BedTool('tests/data/test.bed')
    result_bed = bed_util.extend_bed(
        bed=bed,
        bp=[10, 20],
    )

    assert result_bed[0].chrom == 'chr1'
    assert result_bed[0].start == 90
    assert result_bed[0].end == 220



import pybedtools

def extend_bed(bed, bp, direction='both'):
    """Extend each interval in bed file with specified basepairs.
    If bp is a list, each interval will be extended bp[0] toward upstream, and
    bp[1] toward downstream.

    Note: Be careful! This method does not deal with strandedness of the interval.

    :param (pybedtools.BedTool) bed: Bed object to be manipulated.
    :param (int, list) bp: Each interval will be extended according to bp.
    :param (str) direction: Direction of extension. Note that if bp is a list, direction will be fixed to 'both' ('up', 'down', 'both')
    """
    if isinstance(bp, list):
        assert len(bp) in [1, 2], 'Check parameter `bp`.'
        if len(bp) == 2:
            direction = 'both'
        else:
            bp = [bp[0], bp[0]]
    else:
        bp = [bp, bp]

    upstream_extension = 0 if direction == 'down' else bp[0]
    downstream_extension = 0 if direction == 'up' else bp[1]

    bed_strings = []
    for interval in bed:
        start = max(interval.start - upstream_extension, 0)
        end = interval.end + downstream_extension

        if start < 0:
            continue

        bed_strings.append(
            '%s %d %d %s %s %s' % (interval.chrom, start, end, interval.name, interval.score, interval.strand)
        )

    return pybedtools.BedTool('\n'.join(bed_strings), from_string=True)



from pyensembl import EnsemblRelease

ENSEMBL_RELEASE_VERSION = 87


def gene_by_id(*args, **kwargs):
    genome = EnsemblRelease(ENSEMBL_RELEASE_VERSION)
    return genome.gene_by_id(*args, **kwargs)


def genes_by_name(*args, **kwargs):
    genome = EnsemblRelease(ENSEMBL_RELEASE_VERSION)
    return genome.genes_by_name(*args, **kwargs)


def genes_at_locus(*args, **kwargs):
    genome = EnsemblRelease(ENSEMBL_RELEASE_VERSION)
    return genome.genes_at_locus(*args, **kwargs)


def gene_names_at_locus(*args, **kwargs):
    genome = EnsemblRelease(ENSEMBL_RELEASE_VERSION)
    return genome.gene_names_at_locus(*args, **kwargs)


def symbol2ensg(*args):
    genome = EnsemblRelease(ENSEMBL_RELEASE_VERSION)
    return genome.gene_ids_of_gene_name(*args)[0]



import pysam

import pyBigWig as pbw
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import defaultdict

MAX_DEPTH = 1000000
Y_EXTENTS = [-1000, -5000, -250, -100, -50,
            -25, -10, -5, -2.5, -1, -0.5, 0, 0.5,
            1, 2.5, 5, 10, 25, 50, 100, 250, 500, 1000]

def parse_bam(bam_fp, chrom, start, end):
    """
    """
    data = defaultdict(list)

    bam = pysam.AlignmentFile(bam_fp)
    for pileupcolumn in bam.pileup(chrom, start, end, truncate=True, max_depth=MAX_DEPTH):
        data['chrom'].append(chrom)
        data['pos'].append(pileupcolumn.pos)
        data['coverage'].append(pileupcolumn.n)

    # Assert the dataframe has start and end positions.
    positions = set(data['pos'])
    for pos in [start, end]:
        if pos not in positions:
            data['chrom'].append(chrom) 
            data['pos'].append(pos)
            data['coverage'].append(0)
    
    return pd.DataFrame(data).sort_values('pos').reset_index()

def parse_wiggle(wig_fp, chrom, start, end):
    """
    """
    data = defaultdict(list)
    bw = pbw.open(wig_fp)
    
    values = bw.values(chrom, start, end)
    for pos, v in zip(range(start, end), values):
        if not np.isnan(v) and v != 0:
            data['chrom'].append(chrom)
            data['pos'].append(pos)
            data['value'].append(v)

    positions = set(data['pos'])
    for pos in [start, end]:
        if pos not in positions:
            data['chrom'].append(chrom)
            data['pos'].append(pos)
            data['value'].append(0)

    return pd.DataFrame(data).sort_values('pos').reset_index()

def get_y_extent(v):
    """Automatically determines appropriate y extents of a track.
    """
    v_min, v_max = np.min(v), np.max(v)
    y_min = max([y for y in Y_EXTENTS if y <= v_min])
    y_max = min([y for y in Y_EXTENTS if y >= v_max])

    return y_min, y_max

def format_axis(ax, name, show_coord, ymin, ymax, color=None):
    
    if not show_coord:
        ax.set_xticks([])

    ax.set_ylabel(name, rotation='horizontal', fontdict={'ha': 'right', 'va': 'center', 'fontsize': 10, 'color': ['k', color][color is not None]})
    ax.set_ylim(ymin, ymax)
    ax.set_yticks([ymin, ymax])

    ax.get_yticklabels()[0].set_verticalalignment('bottom')
    ax.get_yticklabels()[1].set_verticalalignment('top')
    ax.get_yticklabels()[0].set_horizontalalignment('right')
    ax.get_yticklabels()[1].set_horizontalalignment('right')

    ax.tick_params(axis='y', direction='in', width=2, labelsize=9)
    ax.spines['left'].set_linewidth(2)
    ax.spines['bottom'].set_linewidth(0.5)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    return ax
   
def get_default_ax():
    fig = plt.figure(figsize=(10, 1))
    ax = fig.add_subplot(111)

    return ax



import re
import tqdm
import pysam
import itertools
import pandas as pd
import numpy as np

from dohlee.genome import Region
from collections import defaultdict, Counter


def get_binary_string_representation(read):
    """Returns a binary representation string of methylation state of a read.
    Unmethylated state = 0, methylated state = 1.

    Requires Bismark-aligned read.
    """
    return ''.join([['0', '1'][c == 'Z'] for c in read.get_tag('XM') if c in ['z', 'Z']])

def character_indices(string, characters):
    i = []
    for character in characters:
        i.extend([m.start() for m in re.finditer(character, string)])
    
    return list(sorted(i))

def get_cpg_coordinates(read, paired=False):
    start = read.reference_start
    methylation_string = read.get_tag('XM')
    
    if not paired:
        if not read.is_reverse:
            return tuple(start + i + 1 for i in character_indices(methylation_string, ['z', 'Z']))
        else:
            return tuple(start + i for i in character_indices(methylation_string, ['z', 'Z']))
        
    else:
        if read.flag in [99, 147]:
            return tuple(start + i + 1 for i in character_indices(methylation_string, ['z', 'Z']))
        else:
            return tuple(start + i for i in character_indices(methylation_string, ['z', 'Z']))
    
def is_reverse(read, paired=False):
    return read.is_reverse if not paired else (read.flag in [99, 147])

def is_concordant(read_bin):
    return len(set(read_bin)) == 1

def pdr(reads_binarized):
    num_discordant = sum(not is_concordant(read_bin) for read_bin in reads_binarized)
    return num_discordant / len(reads_binarized)
    
def methylation_level(reads_binarized):
    num_meth, num_total = 0, 0
    for read_bin in reads_binarized:
        num_meth += sum(CpG == '1' for CpG in read_bin)
        num_total += len(read_bin)
    
    return num_meth / num_total

def methylation_entropy(reads_binarized):
    counts = Counter(reads_binarized)
    b = len(reads_binarized[0])
    N = sum(counts.values())
    return 1/b * sum(-n_i / N * np.log2(n_i / N) for n_i in counts.values())

def epipolymorphism(reads_binarized):
    counts = Counter(reads_binarized)
    N = sum(counts.values())
    ps = [n_i / N for n_i in counts.values()]
    return 1 - sum(p**2 for p in ps)

def get_concordancy_counters1(coordinates, reads_binarized):
    concordant_counter, discordant_counter = Counter(), Counter()
    for read_binarized in reads_binarized:
        indexed_cpgs = [(idx, cpg) for idx, cpg in zip(coordinates, read_binarized)]
        for (idx1, cpg1), (idx2, cpg2) in itertools.combinations(indexed_cpgs, 2):
            dist = abs(idx2 - idx1)
            if cpg1 == cpg2:
                concordant_counter[dist] += 1
            else:
                discordant_counter[dist] += 1

    return concordant_counter, discordant_counter

def get_concordancy_counters2(coordinates, reads_binarized):
    concordant_counter, discordant_counter = Counter(), Counter()
    for read_binarized in reads_binarized:
        indexed_cpgs = [(idx, cpg) for idx, cpg in zip(coordinates, read_binarized)]
        for (idx1, cpg1), (idx2, cpg2) in zip(indexed_cpgs[:-1], indexed_cpgs[1:]):
            dist = abs(idx2 - idx1)
            if cpg1 == cpg2:
                concordant_counter[dist] += 1
            else:
                discordant_counter[dist] += 1

    return concordant_counter, discordant_counter

def is_chr_prepended(path):
    for read in pysam.AlignmentFile(path):
        return read.reference_name.startswith('chr')


def summarize_rrbs(path, file, contigs=None, paired=False, min_CpGs=None, min_depth=None):
    current_region = None
    
    cpgs_reads_dict = defaultdict(list)
    summarized = defaultdict(list)

    bam = pysam.AlignmentFile(path)

    # Setup default contig names if they are not given.
    if contigs is None:
        contigs = list(map(str, range(1, 23))) + ['X', 'Y']
        if is_chr_prepended(path):
            contigs = ['chr' + contig for contig in contigs]

    for contig in contigs:
        count = 0
        
        for read in tqdm.tqdm(bam.fetch(contig), desc=contig):
            read_region = Region(read.reference_name, read.reference_start, read.reference_end)

            if current_region is None:
                current_region = read_region
                cpgs_reads_dict[get_cpg_coordinates(read, paired)].append(read)
                continue

            if current_region.has_overlap_with(read_region):
                current_region = current_region.merge(read_region)
                cpgs_reads_dict[get_cpg_coordinates(read, paired)].append(read)
                
            else:
                # Take the largest bunch of reads.
                best_CpGs = max(cpgs_reads_dict.keys(), key=lambda x: len(cpgs_reads_dict[x]))
                best_reads = cpgs_reads_dict[best_CpGs]
                best_reads_binarized = [get_binary_string_representation(read) for read in best_reads]
                
                # Check if the number of CpGs are all the same.
                assert len(set([len(read_bin) for read_bin in best_reads_binarized])) == 1
                
                meet_cpg = (not min_CpGs) or (min_CpGs and len(best_reads_binarized[0]) >= min_CpGs)
                meet_depth = (not min_depth) or (min_depth and len(best_reads_binarized) >= min_depth)

                # We only deal with RRBS regions with more than 1 CpGs.
                if len(best_reads_binarized[0]) > 0 and meet_cpg and meet_depth:
                    # Summarize.
                    summarized['chrom'].append(current_region.chrom)
                    summarized['CpGs'].append(';'.join([current_region.chrom + ':' + str(CpG) for CpG in best_CpGs]))
                    summarized['num_forward'].append(sum(not is_reverse(read, paired=paired) for read in best_reads))
                    summarized['num_reverse'].append(sum(is_reverse(read, paired=paired) for read in best_reads))
                    summarized['num_concordant'].append(sum(is_concordant(read_bin) for read_bin in best_reads_binarized))
                    summarized['num_discordant'].append(sum(not is_concordant(read_bin) for read_bin in best_reads_binarized))
                    summarized['depth'].append(len(best_reads))
                    summarized['pdr'].append(pdr(best_reads_binarized))
                    summarized['num_CpGs'].append(len(best_reads_binarized[0]))
                    summarized['methylation_level'].append(methylation_level(best_reads_binarized))
                    summarized['methylation_entropy'].append(methylation_entropy(best_reads_binarized))
                    summarized['epipolymorphism'].append(epipolymorphism(best_reads_binarized))
                
                current_region = read_region
                cpgs_reads_dict = defaultdict(list)
                cpgs_reads_dict[get_cpg_coordinates(read, paired)].append(read)
        
    summarized = pd.DataFrame(summarized)
    summarized.to_csv(file, index=False)
    return summarized

def get_concordancy_table(path, file, contigs=None, paired=False, min_depth=10, min_num_cpgs=3):
    current_region = None
    
    cpgs_reads_dict = defaultdict(list)
    summarized = defaultdict(list)

    bam = pysam.AlignmentFile(path)

    # Setup default contig names if they are not given.
    if contigs is None:
        contigs = list(map(str, range(1, 23))) + ['X', 'Y']
        if is_chr_prepended(path):
            contigs = ['chr' + contig for contig in contigs]

    concordant_counter, discordant_counter = Counter(), Counter()
    for contig in contigs:
        count = 0
        
        for read in tqdm.tqdm(bam.fetch(contig), desc=contig):
            read_region = Region(read.reference_name, read.reference_start, read.reference_end)

            if current_region is None:
                current_region = read_region
                cpgs_reads_dict[get_cpg_coordinates(read, paired)].append(read)
                continue

            if current_region.has_overlap_with(read_region):
                current_region = current_region.merge(read_region)
                cpgs_reads_dict[get_cpg_coordinates(read, paired)].append(read)
                
            else:
                # Take the largest bunch of reads.
                best_CpGs = max(cpgs_reads_dict.keys(), key=lambda x: len(cpgs_reads_dict[x]))
                best_reads = cpgs_reads_dict[best_CpGs]
                best_reads_binarized = [get_binary_string_representation(read) for read in best_reads]
                
                # Check if the number of CpGs are all the same.
                assert len(set([len(read_bin) for read_bin in best_reads_binarized])) == 1
                
                # We only deal with RRBS regions with more than or equal to min_depth reads,
                # and more than or equal to min_num_cpgs CpGs.
                depth, num_cpgs = len(best_reads_binarized), len(best_reads_binarized[0])
                if depth >= min_depth and num_cpgs >= min_num_cpgs:
                    # Get concordancy counters.
                    c, d = get_concordancy_counters2(best_CpGs, best_reads_binarized)
                    concordant_counter += c
                    discordant_counter += d
                
                current_region = read_region
                cpgs_reads_dict = defaultdict(list)
                cpgs_reads_dict[get_cpg_coordinates(read, paired)].append(read)

    result = defaultdict(list) 
    for distance in sorted(concordant_counter.keys()):
        c, d = concordant_counter[distance], discordant_counter[distance]
        t = c + d

        result['distance'].append(distance)
        result['mean_concordance'].append(c / t)
        result['num_concordant'].append(c)
        result['num_total'].append(t)
    
    result = pd.DataFrame(result)
    result.to_csv(file, index=False)
    return result



import dohlee.util as util


def given_filename_with_single_dot_split_ext_test():
    filename = '/tmp/a/b/c/d/e/filename.png'
    assert util.strip_ext(filename) == 'filename'


def given_filname_with_multiple_dots_split_ext_test():
    filename = '/tmp/a/b/c/d/e/f.i.l.e.n.a.m.e.png'
    assert util.strip_ext(filename) == 'f.i.l.e.n.a.m.e'



import matplotlib.pyplot as plt

from . import util

def lineplot(data, y, x='pos', name=None, line_kwargs={}, show_coord=False, ax=None):
    """
    """
    if ax is None:
        ax = util.get_default_ax()
    
    sorted_d = data.sort_values(x)
    ax.plot(sorted_d[x].values, sorted_d[y].values, **line_kwargs)
    
    ymin, ymax = util.get_y_extent(sorted_d[y].values)
    ax = util.format_axis(ax, name, show_coord, ymin, ymax)

    return ax

def barplot(data, y, x='pos', name=None, bar_kwargs={}, show_coord=False, ax=None):
    """
    """
    if ax is None:
        ax = util.get_default_ax()

    sorted_d = data.sort_values(x)
    ax.bar(sorted_d[x].values, sorted_d[y].values, width=1.0, lw=0, **bar_kwargs)

    ymin, ymax = util.get_y_extent(sorted_d[y].values)
    ax = util.format_axis(ax, name, show_coord, ymin, ymax)
    
    return ax

def bam(bam_fp, chrom, start, end, name=None, kind='bar', plot_kwargs={}, show_coord=False, ax=None):
    """
    """
    d = util.parse_bam(bam_fp, chrom, start, end)
    if kind == 'bar':
        ax = barplot(d, y='coverage', x='pos', name=name, bar_kwargs=plot_kwargs, show_coord=show_coord, ax=ax)
    elif kind == 'line':
        ax = lineplot(d, y='coverage', x='pos', name=name, line_kwargs=plot_kwargs, show_coord=show_coord, ax=ax)
    
    return ax

def wiggle(wig_fp, chrom, start, end, name=None, kind='line', plot_kwargs={}, show_coord=False, ax=None):
    """
    """
    d = util.parse_wiggle(wig_fp, chrom, start, end)
    if kind == 'line':
        ax = lineplot(d, y='value', x='pos', name=name, line_kwargs=plot_kwargs, show_coord=show_coord, ax=ax)
    elif kind == 'bar':
        ax = barplot(d, y='value', x='pos', name=name, bar_kwargs=plot_kwargs, show_coord=show_coord, ax=ax)
    
    return ax



from collections import namedtuple
import pysam
import numpy as np

class Region(namedtuple('Region', ['chrom', 'start', 'end'])):
    def __new__(cls, chrom, start, end):
        if start >= end:
            raise ValueError('Invalid genomic region %s:%d-%d' % (chrom, start, end))
        
        return super(Region, cls).__new__(cls, chrom, start, end)
    
    def merge(self, r):
        """Merge with another region.
        """
        return merge_regions(self, r)
    
    def has_overlap_with(self, r):
        """Returns True if the region has overlap with r,
        otherwise False.
        """
        return has_overlap(self, r)
    
    def __len__(self):
        return self.end - self.start + 1

    def __str__(self):
        return '%s:%d-%d' % (self.chrom, self.start, self.end)

def merge_regions(r1, r2):
    """Merge two regions.
    """
    assert r1.chrom == r2.chrom, \
        'Unable to merge regions. Contigs are incompatible: %s and %s' % (r1.chrom, r2.chrom)

    start_new, end_new = min(r1.start, r2.start), max(r1.end, r2.end)
    return Region(r1.chrom, start_new, end_new)

def has_overlap(r1, r2):
    if r1.chrom != r2.chrom:
        return False

    a0, a1 = r1.start, r1.end
    b0, b1 = r2.start, r2.end
    
    return a0 <= b1 and b0 <= a1

def fetch_annotation(annotation, chrom, start, end):
    """Returns a generator of annoation rows for given genomic region.
    """
    return pysam.TabixFile(annotation).fetch(chrom, start, end, parser=pysam.asGTF)

def get_gene_ids_at_region(annotation, chrom=None, start=None, end=None, region=None):
    assert (chrom is None) or (region is None), "Please don't specify chrom and region at the same time."
    if region is not None:
        chrom, start, end = region.chrom, region.start, region.end
    fetched = fetch_annotation(annotation, chrom, start, end)
    return [row.gene_id for row in fetched if row.feature == 'gene']

def get_transcript_ids_at_region(annotation, chrom=None, start=None, end=None):
    assert (chrom is None) or (region is None), "Please don't specify chrom and region at the same time."
    if region is not None:
        chrom, start, end = region.chrom, region.start, region.end
    fetched = fetch_annotation(annotation, chrom, start, end)
    return [row.transcript_id for row in fetched if row.feature == 'transcript']

def annotate(table, db):
    if 'chrom' not in table.columns:
        raise ValueError("Column 'chrom' should be specified in the table.")
    if 'start' not in table.columns:
        raise ValueError("Column 'start' should be specified in the table.")
    if 'end' not in table.columns:
        raise ValueError("Column 'end' should be specified in the table.")

    for db_name, db_path in db.items():
        tbx = pysam.TabixFile(db_path)
        values = []
        for record in table.itertuples():
            hits = [read.name for read in tbx.fetch(record.chrom, record.start, record.end, parser=pysam.asBed())]
            # hits = []
            # for read in tbx.fetch(record.chrom, record.start, record.end, parser=pysam.asBed()):
                # if read.start <= record.start and record.end <= read.end:
                    # hits.append(read.name)

            v = ';'.join(hits) if hits else np.nan
            values.append(v)

        table[db_name] = values
    return table

if __name__ == '__main__':
    r1 = Region('chr1', 1000, 2000)
    r2 = Region('chr1', 1500, 3000)
    r3 = Region('chr2', 1000, 2000)
    
    print('len r1 =', len(r1))
    print('len r2 =', len(r2))
    print('len r3 =', len(r3))

    print('Merge(r1, r2) = ', r1.merge(r2))
    print('Merge(r1, r2) = ', merge_regions(r1, r2))

    print('has_overlap(r1, r2) = ', r1.has_overlap_with(r2))
    print('has_overlap(r1, r2) = ', has_overlap(r1, r2))
    print('has_overlap(r1, r3) = ', r1.has_overlap_with(r3))
    print('has_overlap(r1, r3) = ', has_overlap(r1, r3))
    print('has_overlap(r2, r3) = ', r2.has_overlap_with(r3))
    print('has_overlap(r2, r3) = ', has_overlap(r2, r3))



import dohlee.plot as plot; plot.set_style()
import itertools
import tempfile

import numpy as np
import pandas as pd
import seaborn as sns

from collections import Counter
from functools import wraps
from nose.tools import with_setup, raises
from matplotlib.testing.compare import compare_images


def image_comparison(baseline_images, extensions=['png'], tol=1):
    def wrapper(func):
        @wraps(func)
        def decorator(*args, **kwargs):
            results = []
            ax_result = func(*args, **kwargs)
            fp, filename = tempfile.mkstemp(suffix='.png')
            plot.save(filename)

            for baseline_basename in baseline_images:
                for extension in extensions:
                    baseline_image = '.'.join([baseline_basename, extension])

                    result = compare_images(
                        expected=baseline_image,
                        actual=filename,
                        tol=tol,
                    )
        return decorator
    return wrapper


def setup_function():
    pass


def teardown_function():
    plot.clear()


def set_style_test():
    plot.set_style()


@with_setup(setup_function, teardown_function)
def save_test():
    plot.frequency([1, 2, 2, 3, 3, 3])
    plot.save('tmp.png')


@image_comparison(baseline_images=['tests/baseline_images/pca'])
@with_setup(setup_function, teardown_function)
def pca_test():
    iris = sns.load_dataset('iris')
    data = iris[['sepal_length', 'sepal_width', 'petal_length', 'petal_width']]
    plot.pca(data)
    plot.pca(data, title='Dummy title')
    plot.pca(data, labels=iris.species.values, title='Dummy title')


@with_setup(setup_function, teardown_function)
def boxplot_test():
    iris = sns.load_dataset('iris')
    ax = plot.get_axis(transpose=True)
    plot.boxplot(data=iris, x='species', y='sepal_length', ax=ax)


@with_setup(setup_function, teardown_function)
def histogram_test():
    iris = sns.load_dataset('iris')
    ax = plot.get_axis()
    plot.histogram(iris.sepal_length,
                   bins=22,
                   xlabel='Sepal Length',
                   ylabel='Frequency',
                   ax=ax)


@with_setup(setup_function, teardown_function)
def mutation_signature_test():
    data = Counter()
    c_contexts = [p + 'C' + n for (p, n) in itertools.product('ACGT', 'ACGT')]
    t_contexts = [p + 'T' + n for (p, n) in itertools.product('ACGT', 'ACGT')]
    c_alts, t_alts = 'AGT', 'ACG'

    for context, alt in itertools.product(c_contexts, c_alts):
        data[(context, alt)] = np.random.randint(1, 30)
    for context, alt in itertools.product(t_contexts, t_alts):
        data[(context, alt)] = np.random.randint(1, 30)

    ax = plot.get_axis(figsize=(20.4, 3.4))
    plot.mutation_signature(data, ax=ax)
    plot.set_suptitle('Mutational signatures.')


@with_setup(setup_function, teardown_function)
def frequency_test():
    data = [np.random.choice(a=range(10)) for _ in range(100)]
    ax = plot.get_axis()
    plot.frequency(data, ax=ax, xlabel='Your numbers', ylabel='Frequency')


@with_setup(setup_function, teardown_function)
def frequency_sort_by_values_test():
    data = [np.random.choice(a=range(10)) for _ in range(100)]
    ax = plot.get_axis()
    plot.frequency(
        data=data,
        ax=ax,
        sort_by_values=True,
    )


@raises(AssertionError)
@with_setup(setup_function, teardown_function)
def frequency_invalid_order_failing_test():
    data = [1, 1, 3, 3, 4]
    plot.frequency(data=data, order=[1, 2, 3])


@with_setup(setup_function, teardown_function)
def tsne_test():
    iris = sns.load_dataset('iris')
    ax = plot.get_axis()
    plot.tsne(
        iris[['sepal_length', 'sepal_width', 'petal_length', 'petal_width']],
        ax=ax,
        s=5,
        labels=iris['species']
    )


@with_setup(setup_function, teardown_function)
def linear_regression_test():
    x = np.linspace(0, 1, 100)
    y = 2 * x + 3 + np.random.normal(0, 0.3, len(x))

    ax = plot.get_axis()
    plot.linear_regression(x, y, ax=ax)

    ax = plot.get_axis()
    plot.linear_regression(x, y, regression=False, ax=ax)


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_basic_test():
    test_data = pd.DataFrame({
        'Sample': ['S1', 'S2', 'S3'],
        'A': [5, 2, 8],
        'B': [2, 1, 5],
        'C': [1, 4, 2],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'])


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_sort_test():
    test_data = pd.DataFrame({
        'Sample': ['S1', 'S2', 'S3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], sort=True)


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_sort_ascending_test():
    test_data = pd.DataFrame({
        'Sample': ['S1', 'S2', 'S3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], sort=True, reverse=False)


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_single_y_test():
    test_data = pd.DataFrame({
        'Sample': ['S1', 'S2', 'S3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A'])


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_rotate_xticklabels_test():
    test_data = pd.DataFrame({
        'Sample': ['Loooooong label 1', 'Loooooong label 2', 'Loooooong label 3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], rotate_xticklabels=33)


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_hide_xticklabels_test():
    test_data = pd.DataFrame({
        'Sample': ['Loooooong label 1', 'Loooooong label 2', 'Loooooong label 3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], xticklabels=False)


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_sort_by_test():
    test_data = pd.DataFrame({
        'Sample': ['S1', 'S2', 'S3', 'S4', 'S5'],
        'A': [1, 3, 5, 1, 5],
        'B': [2, 4, 1, 2, 5],
        'C': [3, 8, 4, 6, 3],
        'Group': ['G1', 'G2', 'G1', 'G1', 'G2'],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], sort=True, sort_by='B')


@with_setup(setup_function, teardown_function)
def stacked_bar_chart_group_test():
    test_data = pd.DataFrame({
        'Sample': ['S1', 'S2', 'S3', 'S4', 'S5'],
        'A': [1, 3, 5, 1, 5],
        'B': [2, 4, 1, 2, 5],
        'C': [3, 8, 4, 6, 3],
        'Group': ['G1', 'G2', 'G1', 'G1', 'G2'],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], group='Group')


@with_setup(setup_function, teardown_function)
def legend_size_small_test():
    test_data = pd.DataFrame({
        'Sample': ['Loooooong label 1', 'Loooooong label 2', 'Loooooong label 3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], legend_size='small')


@with_setup(setup_function, teardown_function)
def legend_size_medium_test():
    test_data = pd.DataFrame({
        'Sample': ['Loooooong label 1', 'Loooooong label 2', 'Loooooong label 3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], legend_size='medium')


@with_setup(setup_function, teardown_function)
def legend_size_large_test():
    test_data = pd.DataFrame({
        'Sample': ['Loooooong label 1', 'Loooooong label 2', 'Loooooong label 3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], legend_size='large')


@with_setup(setup_function, teardown_function)
def legend_title_test():
    test_data = pd.DataFrame({
        'Sample': ['Loooooong label 1', 'Loooooong label 2', 'Loooooong label 3'],
        'A': [1, 3, 5],
        'B': [2, 4, 1],
        'C': [3, 8, 4],
    })
    plot.stacked_bar_chart(data=test_data, x='Sample', y=['A', 'B', 'C'], legend_title='Mutation type')


@with_setup(setup_function, teardown_function)
def umap_test():
    iris = sns.load_dataset('iris')
    ax = plot.get_axis()
    plot.umap(
        iris[['sepal_length', 'sepal_width', 'petal_length', 'petal_width']],
        ax=ax,
        s=5,
        labels=iris['species']
    )


@with_setup(setup_function, teardown_function)
def pca_with_no_xyticklabels_test():
    iris = sns.load_dataset('iris')
    ax = plot.get_axis(figsize=(8.4, 8.4))
    plot.pca(
        data=iris[['sepal_length', 'sepal_width', 'petal_length', 'petal_width']],
        labels=iris.species,
        s=5,
        ax=ax,
        xticklabels=False,
        yticklabels=False,
    )


@with_setup(setup_function, teardown_function)
def dimensionality_reduction_test():
    iris = sns.load_dataset('iris')
    ax = plot.get_axis(figsize=(8.4 * 3, 8.4))
    plot.dimensionality_reduction(
        data=iris[['sepal_length', 'sepal_width', 'petal_length', 'petal_width']],
        labels=iris.species,
        ax=ax,
        xticklabels=False,
        yticklabels=False,
    )


@with_setup(setup_function, teardown_function)
def line_test():
    fmri = sns.load_dataset('fmri')
    ax = plot.get_axis()
    plot.line(
        data=fmri,
        x='timepoint',
        y='signal',
    )
